package GD.Data;

/**
 * Eye enumeration. This enumeration was made to unify the values that are used
 * for storing eye identification values in {@link JGDFixation JGDFixation} and
 * {@link JGDSingleSample JGDSingleSample}.
 * 
 * @version 2.0.0
 * @author Andrey Shipilov
 */
public enum EEye {
	/** Left eye */
	LEFT,

	/** Right eye */
	RIGHT,

	/** Unknown. This value is used in case if input cannot be parsed */
	UNKNOWN;

	/**
	 * Parse input EGDEye enumeration value into integer representation
	 * 
	 * @param e
	 *            - EGDEye value, representing eye
	 * @return integer value, representing eye
	 * 
	 * @see EEye#parse(int)
	 */
	public static int parse(EEye e) {
		switch (e) {
		case LEFT:
			return 0;
		case RIGHT:
			return 1;
		default:
			return 999;
		}
	}

	/**
	 * Parse input integer value trying to identify the eye
	 * 
	 * @param i
	 *            - integer value, representing eye
	 * @return EEye enumeration value
	 * 
	 * @see EEye#parse(EEye)
	 */
	public static EEye parse(int i) {
		switch (i) {
		case 0:
			return LEFT;
		case 1:
			return RIGHT;
		default:
			return UNKNOWN;
		}
	}
}
