package GD.Data;

import java.util.Vector;

public class GazeContainer implements IGazeContainer {

    /**
     * Default length of the queue of blinks
     */
    public static int DEFAULT_BLINKS_LENGTH = 10;
    /**
     * Default length of the queue of fixations
     */
    public static int DEFAULT_FIXATIONS_LENGTH = 10;
    /**
     * Default length of the queue of samples
     */
    public static int DEFAULT_SAMPLES_LENGTH = 10;
    /**
     * Blinks queue
     */
    protected Queue<GazeEntryBlink> mBlinks;
    /**
     * Fixations queue
     */
    protected Queue<GazeEntryFixation> mFixations;
    /**
     * Right eye samples queue
     */
    protected Queue<GazeEntrySample> mSamples;

    public GazeContainer() {
        mBlinks = new Queue<GazeEntryBlink>(DEFAULT_BLINKS_LENGTH);
        mFixations = new Queue<GazeEntryFixation>(DEFAULT_FIXATIONS_LENGTH);
        mSamples = new Queue<GazeEntrySample>(DEFAULT_SAMPLES_LENGTH);
    }

    public GazeContainer(int blinks, int fixations, int samples) {
        if (blinks < 1) {
            blinks = DEFAULT_BLINKS_LENGTH;
        }
        if (fixations < 1) {
            fixations = DEFAULT_FIXATIONS_LENGTH;
        }
        if (samples < 1) {
            samples = DEFAULT_SAMPLES_LENGTH;
        }

        mBlinks = new Queue<GazeEntryBlink>(blinks);
        mFixations = new Queue<GazeEntryFixation>(fixations);
        mSamples = new Queue<GazeEntrySample>(samples);
    }

    @Override
    public synchronized <T extends GazeEntryBase> void Add(T entry) {
        //String msg = entry.ToLogString();
        switch (entry.Type()) {
            case ET_BLINK:
                mBlinks.append((GazeEntryBlink) entry);
                break;
            case ET_FIXATION:
                mFixations.append((GazeEntryFixation) entry);
                break;
            case ET_SAMPLE:
                mSamples.append((GazeEntrySample) entry);
                break;
            default:
               // msg = entry.toString();
                break;
        }
        
       // System.out.println(msg);
    }

    @SuppressWarnings("unchecked")
    @Override
    public synchronized <T extends GazeEntryBase> Queue<T> Get(EEntryType type)
            throws Exception {
        switch (type) {
            case ET_BLINK:
                return (Queue<T>) mBlinks;
            case ET_FIXATION:
                return (Queue<T>) mFixations;
            case ET_SAMPLE:
                return (Queue<T>) mSamples;
        }

        throw new Exception("Not supported type: " + type.toString());
    }

    @Override
    public <T extends GazeEntryBase> void Add(Vector<T> entries) {
        for(int i = 0; i < entries.size(); i++)
            Add(entries.elementAt(i));
    }
}
