package GD.Data;

import java.awt.Point;

/**
 * Single sample gaze data. This gaze data contains coordinates, so it cannot be
 * tracked against screen coordinate system and drawn on the screen.
 * 
 * @see IGDEntry
 * @see IGDTrackableEntry
 * @see JGDFixation
 * @see JGDBlink
 * 
 * @version 2.1.0
 * @author Andrey Shipilov
 * 
 */
public class GazeEntrySample extends GazeEntryBase implements ICoordinates {
    public static final int MSG_SIZE = 24;
    
	/** Eye which generated sample event */
	EEye mEye;

	/** Size of the pupil at the moment of catching this gaze data by tracker */
	int mPupilSize;

	/** Event id, assigned by the tracker */
	int mEvent;

	/** Coordinates of the gaze. It is stored in server coordinate system */
	Point mCoords;

	/**
	 * Constructor
	 * 
	 * @param ordinal
	 *            Ordinal number of sample
	 * @param time
	 *            Time stamp of sample
	 * @param event
	 *            Event id of the sample
	 * @param eye
	 *            Eye, generated sample event
	 * @param pupil
	 *            Pupil size
	 * @param coords
	 *            Coordinates of the gaze
	 */
	public GazeEntrySample(int ordinal, int time, int event, int eye,
			int pupil, Point coords) {
		super(ordinal, time);

		mPupilSize = pupil;
		mEvent = event;
		mEye = EEye.parse(eye);

		mCoords = coords;
	}

	@Override
	public int MessageLength() {
		return MSG_SIZE;
	}

	@Override
	public byte[] MessageStruct() {
		byte[] msg = new byte[MessageLength()];

                msg[0] = (byte) (mOrdinal & 0xff);
                msg[1] = (byte) ((mOrdinal >> 8) & 0xff);
                msg[2] = (byte) ((mOrdinal >> 16) & 0xff);
		msg[3] = (byte) ((mOrdinal >> 24) & 0xff);

                msg[4] = (byte) (mTimestamp & 0xff);
                msg[5] = (byte) ((mTimestamp >> 8) & 0xff);
                msg[6] = (byte) ((mTimestamp >> 16) & 0xff);
		msg[7] = (byte) ((mTimestamp >> 24) & 0xff);

		int ieye = EEye.parse(mEye);
                msg[8] = (byte) (ieye & 0xff);
                msg[9] = (byte) ((ieye >> 8) & 0xff);
                msg[10] = 0;
                msg[11] = 0;

                msg[12] = (byte) (mCoords.x & 0xff);
                msg[13] = (byte) ((mCoords.x >> 8) & 0xff);
                msg[14] = (byte) ((mCoords.x >> 16) & 0xff);
		msg[15] = (byte) ((mCoords.x >> 24) & 0xff);
		
		msg[16] = (byte) (mCoords.y & 0xff);
		msg[17] = (byte) ((mCoords.y >> 8) & 0xff);
                msg[18] = (byte) ((mCoords.y >> 16) & 0xff);
                msg[19] = (byte) ((mCoords.y >> 24) & 0xff);

		msg[20] = (byte) (mPupilSize & 0xff);
		msg[21] = (byte) ((mPupilSize >> 8) & 0xff);
                msg[22] = (byte) ((mPupilSize >> 16) & 0xff);
                msg[23] = (byte) ((mPupilSize >> 24) & 0xff);

		return msg;
	}

	@Override
	public EEntryType Type() {
		return EEntryType.ET_SAMPLE;
	}

	@Override
	public Point Coords() {
		return mCoords;
	}

	@Override
	/**
	 * Format :
	 * 		Timestamp	Type	Ordinal	Duration	Eye	PupilSize	Event X Y
	 */
	public String ToLogString() {
            String str = "";
            str += String.valueOf(mTimestamp) + "\t";
            str += String.valueOf(Type()) + "\t";
            str += String.valueOf(mOrdinal) + "\t";
            str += String.valueOf("-") + "\t";
            str += String.valueOf(mEye) + "\t";
            str += String.valueOf(mPupilSize) + "\t";
            str += String.valueOf(mEvent) + "\t";
            str += String.valueOf(mCoords.x) + "\t";
            str += String.valueOf(mCoords.y);
            
            return str;
	}

}