package GD.Graphics;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Iterator;

import GD.Data.EEntryType;
import GD.Data.GazeClient;
import GD.Data.IGazeClient;
import GD.Data.IGazeContainer;
import java.awt.Color;

abstract class GazeMapProducer implements IGazeMapProducer{
	protected boolean mDrawShapes;
	protected boolean mDrawLabels;
	protected GazeMarkBase mMark;
	protected EEntryType mDrawMode;
        

	public GazeMapProducer(GazeMarkBase mark) throws Exception {
            if (mark == null)
                throw new Exception("Mark cannot be null");
            
            mMark = mark;
	}

	@Override
	public GazeMarkBase getMark() {
		return mMark;
	}
	
	public void setMark(GazeMarkBase mark) {
		mMark = mark;
	}

	public boolean getDrawShapes() {
		return mDrawShapes;
	}

	public void setDrawShapes(boolean drawShapes) {
		this.mDrawShapes = drawShapes;
	}
	
	public boolean getDrawLabels() {
		return mDrawLabels;
	}

	public void setDrawLabels(boolean drawLabels) {
		this.mDrawLabels = drawLabels;
	}
	
	public EEntryType getDrawMode() {
		return mDrawMode;
	}

	public void setDrawMode(EEntryType drawMode) {
		this.mDrawMode = drawMode;
	}
        
        private void clear(BufferedImage image){
            Graphics g = image.getGraphics();
            Graphics2D g2d = (Graphics2D)g;
            g2d.setBackground(new Color(0,0,0,0));
            //g.setColor(new Color(254,0,0,254));
            g.clearRect(0, 0, image.getWidth(), image.getHeight());
        }

	public Rectangle2D BuildMap(HashMap<Short, IGazeClient> clients,
			BufferedImage image) {
		Rectangle2D updateRect = new Rectangle();
		Graphics g = image.getGraphics();
                
                clear(image);

		Iterator<IGazeClient> iter = clients.values().iterator();
		while (iter.hasNext()) {
			IGazeClient client = iter.next();
			if (client.getVisibility()) {
				g.setColor(client.getColor());

				if (mDrawShapes) {
                                    IGazeContainer cnt = client.GazeContainer();
                                    if (cnt != null){
                                        synchronized(cnt){
                                            Rectangle2D r = drawShape(g, cnt);
                                            if (!r.isEmpty())
                                                    updateRect.add(r);
                                        }
                                    }
				}

				if (mDrawLabels && g instanceof Graphics2D) {
                                    String name = client.getName();
                                    if (name != null){
                                        synchronized(name){
                                            Rectangle2D r = drawLabel((Graphics2D) g,
                                                            client.GazeContainer(), name);
                                            if (!r.isEmpty())
                                                    updateRect.add(r);
                                        }
                                    }
				}
			}
		}

		return updateRect;
	}

	abstract protected Rectangle2D drawShape(Graphics g, IGazeContainer gazes);

	abstract protected Rectangle2D drawLabel(Graphics2D g,
			IGazeContainer gazes, String label);
}
