package GazeVnc;

import GD.Connection.ETUD.ETUDGazeSinkThread;
import GD.Connection.ETUD.ETUDGazeSource;
import GD.Connection.GazeSourceFactory;
import GD.Connection.IGazeSource;
import GD.Data.*;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import GD.GUI.IGazeCanvas;
import GD.Graphics.*;
import TightVnc.RfbProto;
import TightVnc.VncCanvas;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.MemoryImageSource;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

public class GazeVncCanvas extends VncCanvas
        implements ICoordinateConverter, IGazeCanvas {

    /*Static class members*/
    public static String DEFAULT_NAME = "<none>";
    public static String DEFAULT_DESCRIPTION = "<none>";
    public static short SELF_ID = 0;
    public static int DEFAULT_FPS = 25;
    public static int DEFAULT_MPS = 25;
    
    protected GazeVncViewer mViewer;
    final protected HashMap<Short, IGazeClient> mClients;
    
    /* Self information */
    protected float mScale;
    protected String mName;
    protected String mDescription;
    
    /* Gaze map */
    protected IGazeMapProducer mGazeMapProducer;
    protected int mFPS;
    protected int mGazeMapRefreshPeriod;
    protected Timer mGazeMapDrawTimer;
    protected GazeMapRefreshTask mGazeDrawTask;
    protected BufferedImage mGazeMap;
    protected BufferedImage mOutputImage;
    protected Graphics mOutputImageGraphics;
    protected boolean mGazeMapDrawTimerOn;
    
    /*Data processing*/
    protected IGazeContainer mGazeContainer;
    protected IGazeSource mGazeSource;
    protected ETUDGazeSinkThread mGazeSourceThread;
    protected GazeDataSendTask mGazeSendTask;
    protected int mMPS;
    protected int mGazeDataSendPeriod;
    protected Timer mGazeDataSendTimer;
    protected boolean mGazeDataSendTimerOn;
    
    /*Constructor*/
    public GazeVncCanvas(GazeVncViewer v, int maxWidth, int maxHeight) throws IOException, Exception {
        super(v, maxWidth, maxHeight);

        mViewer = v;
        mClients = new HashMap<Short, IGazeClient>();
        mGazeContainer = new GazeContainer(); // new DebugGazeContainer();
        mGazeMapProducer = new LastGazeMapProducer(GazeMarkFactory.Products[0]);
        mGazeMapProducer.setDrawLabels(true);
        mGazeMapProducer.setDrawShapes(true);
        mGazeMapProducer.setDrawMode(EEntryType.ET_SAMPLE);

        mName = DEFAULT_NAME;
        mDescription = DEFAULT_DESCRIPTION;

        mScale = 100;
        Converter.initialize(this);

        mGazeMapDrawTimer = new Timer();
        mGazeDataSendTimer = new Timer();
        
        mGazeDrawTask = new GazeMapRefreshTask();
        mGazeSendTask = new GazeDataSendTask();
        
        mGazeDataSendTimerOn = false;
        mGazeMapDrawTimerOn = false;
                
        InitGazeSinkSystem();
        
        setMPS(DEFAULT_MPS);
        setFPS(DEFAULT_FPS);

        return;
    }

    @Override
    public void StartTracking() {
        mGazeSource.StartTracking();
        try {
              mViewer.mGazeProto.WriteRegisterMsg();
        } catch (IOException ex) {
               Logger.getLogger(GazeVncCanvas.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public void StopTracking() {
        mGazeSource.StopTracking();
        try {
            mViewer.mGazeProto.WriteUnregisterMsg();
        } catch (IOException ex) {
              Logger.getLogger(GazeVncCanvas.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public int getMPS() {
        return mMPS;
    }

    @Override
    public void setMPS(int mps) {
        mMPS = mps;
        if (mMPS < 5 || mMPS > 60)
            mMPS = 25;
        
        if (mGazeSendTask != null)
            mGazeSendTask.cancel();
                
        mGazeSendTask = new GazeDataSendTask();
        mGazeDataSendTimer.scheduleAtFixedRate(mGazeSendTask, 0, (long)(1000 / mMPS));
        mGazeDataSendTimerOn = true;
    }

    @Override
    public int getFPS() {
        return mFPS;
    }

    @Override
    public void setFPS(int fps) {
        mFPS = fps;
        if (mFPS < 5 || mFPS > 60)
            mFPS = 25;
        
        if (mGazeDrawTask != null)
            mGazeDrawTask.cancel();
        
        mGazeDrawTask = new GazeMapRefreshTask();
        mGazeMapDrawTimer.scheduleAtFixedRate(mGazeDrawTask, 0, (long)(1000 / mFPS));
        mGazeMapDrawTimerOn = true;
    }

    class GazeMapRefreshTask extends TimerTask {
        Rectangle2D prevRect;

        public GazeMapRefreshTask() {
            prevRect = new Rectangle();
        }

        @Override
        public void run() {
            if (mGazeMap != null) {
                synchronized (mGazeMap) {
                    synchronized (mClients) {
                        Rectangle2D updRect = mGazeMapProducer.BuildMap(mClients, mGazeMap);

                        prevRect.add(updRect);
                        scheduleRepaint((int) prevRect.getMinX(), (int) prevRect.getMinY(), (int) prevRect.getWidth(), (int) prevRect.getHeight());
                        prevRect.setRect(updRect);
                    }
                }
            }
        }
    }

    class GazeDataSendTask extends TimerTask {
        public void run() {
            if (mGazeContainer != null) {
                synchronized (mGazeContainer) {
                    try {
                        mViewer.mGazeProto.WriteSamplesMsg(mGazeContainer.Get(EEntryType.ET_SAMPLE));
                        mViewer.mGazeProto.WriteFixationsMsg(mGazeContainer.Get(EEntryType.ET_FIXATION));
                        mViewer.mGazeProto.WriteBlinksMsg(mGazeContainer.Get(EEntryType.ET_BLINK));
                    } catch (Exception ex) {
                        Logger.getLogger(GazeVncCanvas.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }
            }
        }
    }

    public void InitGazeSinkSystem() {
        if (mGazeSource == null)
            mGazeSource = new ETUDGazeSource(mGazeContainer);

        if (mGazeSourceThread == null && mGazeSource != null)
            mGazeSourceThread = new ETUDGazeSinkThread((mGazeSource));

        if (mGazeSourceThread != null && !mGazeSourceThread.isAlive())
            mGazeSourceThread.start();
    }

    @Override
    public int XOffset() {
        return (int) getLocationOnScreen().x;
    }

    @Override
    public double XScale() {
        return mScale;
    }

    @Override
    public int YOffset() {
        return (int) getLocationOnScreen().y;
    }

    @Override
    public double YScale() {
        return mScale;
    }
    
    @Override
    public int Width() {
        return getWidth();
    }
    
    @Override
    public int Height() {
        return getHeight();
    }

    @Override
    public IGazeClient getClient(int id) {
        if (!mClients.containsKey(id)) {
            return null;
        }

        return mClients.get(id);
    }

    @Override
    public HashMap<Short, IGazeClient> Clients() {
        return mClients;
    }

    @Override
    public IGazeMapProducer MapProducer() {
        return mGazeMapProducer;
    }

    @Override
    public IGazeContainer GazeContainer() {
        return mGazeContainer;
    }

    @Override
    public String getDescription() {
        return mDescription;
    }

    @Override
    public void setDescription(String description) {
        try {
            if (description.length() > GazeClient.MAX_DESCRIPTION_LENGTH)
                description = description.substring(0, GazeClient.MAX_DESCRIPTION_LENGTH);

            mDescription = description;
            mViewer.mGazeProto.WriteDescriptionMsg(mDescription);
        } catch (IOException ex) {
            Logger.getLogger(GazeVncCanvas.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public String getName() {
        return mName;
    }

    @Override
    public void setName(String name) {
        try {
            if (name.length() > GazeClient.MAX_NAME_LENGTH)
                name = name.substring(0, GazeClient.MAX_NAME_LENGTH);
            
            mName = name;
            mViewer.mGazeProto.WriteNameMsg(mName);
        } catch (IOException ex) {
            Logger.getLogger(GazeVncCanvas.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public IGazeSource GazeSource() {
        return mGazeSource;
    }

    @Override
    public void paint(Graphics g) {
        synchronized (mOutputImage) {
            mOutputImageGraphics.clearRect(0, 0, mOutputImage.getWidth(), mOutputImage.getHeight());

            synchronized (memImage) {
                mOutputImageGraphics.drawImage(memImage, 0, 0, null);
            }
            synchronized (mGazeMap) {
                mOutputImageGraphics.drawImage(mGazeMap, 0, 0, null);
            }

            if (rfb.framebufferWidth == scaledWidth) {
                g.drawImage(mOutputImage, 0, 0, null);
            } else {
                paintScaledFrameBuffer(g);
            }
        }

        if (showSoftCursor) {
            int x0 = cursorX - hotX, y0 = cursorY - hotY;
            Rectangle r = new Rectangle(x0, y0, cursorWidth, cursorHeight);
            if (r.intersects(g.getClipBounds())) {
                g.drawImage(softCursor, x0, y0, null);
            }
        }
    }

    public void paintScaledFrameBuffer(Graphics g) {
        g.drawImage(mOutputImage, 0, 0, scaledWidth, scaledHeight, null);
    }

    @Override
    public void processNormalProtocol() throws Exception {

        // Start/stop session recording if necessary.
        viewer.checkRecordingStatus();

        rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
                rfb.framebufferHeight, false);

        resetStats();
        boolean statsRestarted = false;

        //
        // main dispatch loop
        //

        while (true) {

            // Read message type from the server.
            int msgType = rfb.readServerMessageType();
            
            // Process the message depending on its type.
            switch (msgType) {
                case RfbProto.FramebufferUpdate:

                    if (statNumUpdates == viewer.debugStatsExcludeUpdates
                            && !statsRestarted) {
                        resetStats();
                        statsRestarted = true;
                    } else if (statNumUpdates == viewer.debugStatsMeasureUpdates
                            && statsRestarted) {
                        viewer.disconnect();
                    }

                    rfb.readFramebufferUpdate();
                    statNumUpdates++;

                    boolean cursorPosReceived = false;

                    for (int i = 0; i < rfb.updateNRects; i++) {

                        rfb.readFramebufferUpdateRectHdr();
                        statNumTotalRects++;
                        int rx = rfb.updateRectX, ry = rfb.updateRectY;
                        int rw = rfb.updateRectW, rh = rfb.updateRectH;

                        if (rfb.updateRectEncoding == RfbProto.EncodingLastRect) {
                            break;
                        }

                        if (rfb.updateRectEncoding == RfbProto.EncodingNewFBSize) {
                            rfb.setFramebufferSize(rw, rh);
                            updateFramebufferSize();
                            break;
                        }

                        if (rfb.updateRectEncoding == RfbProto.EncodingXCursor
                                || rfb.updateRectEncoding == RfbProto.EncodingRichCursor) {
                            handleCursorShapeUpdate(rfb.updateRectEncoding, rx, ry,
                                    rw, rh);
                            continue;
                        }

                        if (rfb.updateRectEncoding == RfbProto.EncodingPointerPos) {
                            softCursorMove(rx, ry);
                            cursorPosReceived = true;
                            continue;
                        }

                        long numBytesReadBefore = rfb.getNumBytesRead();

                        rfb.startTiming();

                        switch (rfb.updateRectEncoding) {
                            case RfbProto.EncodingRaw:
                                statNumRectsRaw++;
                                handleRawRect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingCopyRect:
                                statNumRectsCopy++;
                                handleCopyRect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingRRE:
                                handleRRERect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingCoRRE:
                                handleCoRRERect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingHextile:
                                statNumRectsHextile++;
                                handleHextileRect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingZRLE:
                                statNumRectsZRLE++;
                                handleZRLERect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingZlib:
                                handleZlibRect(rx, ry, rw, rh);
                                break;
                            case RfbProto.EncodingTight:
                                statNumRectsTight++;
                                handleTightRect(rx, ry, rw, rh);
                                break;
                            default:
                                throw new Exception("Unknown RFB rectangle encoding "
                                        + rfb.updateRectEncoding);
                        }

                        rfb.stopTiming();

                        statNumPixelRects++;
                        statNumBytesDecoded += rw * rh * bytesPixel;
                        statNumBytesEncoded += (int) (rfb.getNumBytesRead() - numBytesReadBefore);
                    }

                    boolean fullUpdateNeeded = false;

                    // Start/stop session recording if necessary. Request full
                    // update if a new session file was opened.
                    if (viewer.checkRecordingStatus()) {
                        fullUpdateNeeded = true;
                    }

                    // Defer framebuffer update request if necessary. But wake up
                    // immediately on keyboard or mouse event. Also, don't sleep
                    // if there is some data to receive, or if the last update
                    // included a PointerPos message.
                    if (viewer.deferUpdateRequests > 0 && rfb.available() == 0
                            && !cursorPosReceived) {
                        synchronized (rfb) {
                            try {
                                rfb.wait(viewer.deferUpdateRequests);
                            } catch (InterruptedException e) {
                            }
                        }
                    }

                    viewer.autoSelectEncodings();

                    // Before requesting framebuffer update, check if the pixel
                    // format should be changed.
                    if (viewer.options.eightBitColors != (bytesPixel == 1)) {
                        // Pixel format should be changed.
                        setPixelFormat();
                        fullUpdateNeeded = true;
                    }

                    // Request framebuffer update if needed.
                    int w = rfb.framebufferWidth;
                    int h = rfb.framebufferHeight;
                    rfb.writeFramebufferUpdateRequest(0, 0, w, h, !fullUpdateNeeded);

                    break;

                case RfbProto.SetColourMapEntries:
                    throw new Exception("Can't handle SetColourMapEntries message");

                case RfbProto.Bell:
                    Toolkit.getDefaultToolkit().beep();
                    break;

                case RfbProto.ServerCutText:
                    String s = rfb.readServerCutText();
                    viewer.clipboard.setCutText(s);
                    break;
                    
                case GazeProto.gdControlMsg:
                    synchronized(mViewer.mGazeProto.inlock){
                        int subtype = mViewer.mGazeProto.readServerMessageSubtype();
                        switch(subtype)
                        {
                            case GazeProto.gdPoke:
                                short pokeId = mViewer.mGazeProto.ReadPokeMsg();
                                if (mClients.containsKey(pokeId))
                                    System.out.println("Poke client with id " + pokeId);
                                break;
                            case GazeProto.gdRegister:
                                short regId = mViewer.mGazeProto.ReadPokeMsg();
                                if (!mClients.containsKey(regId))
                                    mClients.put(regId, new GazeClient(regId));
                                break;
                            case GazeProto.gdUnregister:
                                short unRegId = mViewer.mGazeProto.ReadPokeMsg();
                                if (mClients.containsKey(unRegId))
                                    mClients.remove(unRegId);
                                break;
                            default:
                                break;
                        }
                    }
                    break;
                   
                case GazeProto.gdTextMsg:
                    synchronized(mViewer.mGazeProto.inlock){
                        int subtype = mViewer.mGazeProto.readServerMessageSubtype();
                        switch(subtype)
                        {
                            case GazeProto.gdName:
                                String _name = mViewer.mGazeProto.ReadNameMsg();
                                short nameId = mViewer.mGazeProto.srcClientId;
                                if (mClients.containsKey(nameId))
                                    mClients.get(nameId).setName(_name);
                                break;
                            case GazeProto.gdDescription:
                                String _desc = mViewer.mGazeProto.ReadDescriptionMsg();
                                short descId = mViewer.mGazeProto.srcClientId;
                                if (mClients.containsKey(descId))
                                    mClients.get(descId).setDescription(_desc);                                break;
                            case GazeProto.gdText:
                                String _text = mViewer.mGazeProto.ReadTextMsg();
                                short textId = mViewer.mGazeProto.srcClientId;
                                if (mClients.containsKey(textId))
                                    System.out.println("Received message from client " + textId + ". (" + _text + ")");
                                break;
                            default:
                                break;
                        }
                    }
                    break;
                    
                case GazeProto.gdGazeMsg:
                    synchronized(mViewer.mGazeProto.inlock){
                        int subtype = mViewer.mGazeProto.readServerMessageSubtype();
                        switch(subtype)
                        {
                            case GazeProto.gdGazeBlink:
                                Vector<GazeEntryBlink> b = mViewer.mGazeProto.ReadBlinksMsg();
                                short blinksId = mViewer.mGazeProto.srcClientId;
                                if (mClients.containsKey(blinksId))
                                    mClients.get(blinksId).GazeContainer().Add(b);
                                break;
                            case GazeProto.gdGazeFixation:
                                Vector<GazeEntryFixation> f = mViewer.mGazeProto.ReadFixationsMsg();
                                short fixsId = mViewer.mGazeProto.srcClientId;
                                if (mClients.containsKey(fixsId))
                                    mClients.get(fixsId).GazeContainer().Add(f);
                                break;
                            case GazeProto.gdGazeSample:
                                Vector<GazeEntrySample> ss = mViewer.mGazeProto.ReadSamplesMsg();
                                short samplesId = mViewer.mGazeProto.srcClientId;
                                if (mClients.containsKey(samplesId))
                                    mClients.get(samplesId).GazeContainer().Add(ss);
                                break;
                            default:
                                break;
                        }
                    }
                    break;

                default:
                    throw new Exception("Unknown RFB message type " + msgType);
            }
        }
    }

    @Override
    protected void updateFramebufferSize() {

        // Useful shortcuts.
        int fbWidth = rfb.framebufferWidth;
        int fbHeight = rfb.framebufferHeight;

        if (mOutputImage == null) {
            mOutputImage = new BufferedImage(fbWidth, fbHeight, BufferedImage.TYPE_INT_ARGB);
            mOutputImageGraphics = mOutputImage.getGraphics();
        } else if (mOutputImage.getWidth(null) != fbWidth
                || mOutputImage.getHeight(null) != fbHeight) {
            synchronized (mOutputImage) {
                mOutputImage = new BufferedImage(fbWidth, fbHeight, BufferedImage.TYPE_INT_ARGB);
                mOutputImageGraphics = mOutputImage.getGraphics();
            }
        }


        if (mGazeMap == null) {
            mGazeMap = new BufferedImage(fbWidth, fbHeight, BufferedImage.TYPE_INT_ARGB);

        } else if (mGazeMap.getWidth(null) != fbWidth
                || mGazeMap.getHeight(null) != fbHeight) {
            synchronized (mGazeMap) {
                mGazeMap = new BufferedImage(fbWidth, fbHeight, BufferedImage.TYPE_INT_ARGB);
            }
        }

        super.updateFramebufferSize();
    }
}
